"""
Main pipeline demonstrating how to use SQUID with example models
"""

import os, sys
sys.dont_write_bytecode = True
import numpy as np
import mutagenizer
import predictor
import mave
import surrogate_zoo
import impress
py_dir = os.path.dirname(os.path.abspath(__file__))


# =============================================================================
# Computational settings
# =============================================================================
gpu = False
save = True


# =============================================================================
# Setup for user-defined deep learning model (choose one)
# =============================================================================
if 1:
    # import ResidualBind-32 model and define hyperparameters
    model_name = 'ResidualBind32'
    from example_models.ResidualBind32 import re32_utils
    model, bin_size = re32_utils.read_model(os.path.join(py_dir,'example_models/%s/model' % model_name), compile_model=True)
    """
        Inputs shape:   (n, 2048, 4)
        Outputs shape:  (1, 64, 15) : 64-bin-resolution profile for each of the 15 cell lines
    """

    # define sequence-of-interest and output head (task)
    seq_index = 641
    task_idx = 13 #cell line

    # create save directory
    if save is True:
        save_dir = os.path.join(py_dir, 'squid_outputs/%s/%s' % (model_name, seq_index))
        if not os.path.exists(save_dir):
            os.makedirs(save_dir)
    else:
        save_dir = None

    # retrieve sequence-of-interest from test set
    import h5py
    with h5py.File(os.path.join(py_dir, 'example_models/%s/cell_line_%s.h5' % (model_name, task_idx)), 'r') as dataset:
        x_all = np.array(dataset['X']).astype(np.float32)

    # define mutagenesis window for sequence
    start_position = 962-10 #e.g., the nucleotide at position 962 is the start of a pronounced AP1 motif (TGANTCA) in cell line 13
    stop_position = 962+7+10

    # set up predictor class for in silico MAVE
    pred_generator = predictor.ProfilePredictor(pred_fun=model.predict_on_batch, task_idx=task_idx,
                                                batch_size=512, save_dir=save_dir,
                                                reduce_fun=predictor.profile_sum,
                                                #reduce_fun=predictor.profile_pca,
                                                )
    
    alphabet = ['A','C','G','T']
    log2FC = False
    output_skip = 0
    

elif 0:
    # import DeepSTARR model and define hyperparameters
    model_name = 'DeepSTARR'

    def load_model(model):
        from keras.models import model_from_json
        keras_model_weights = model + '.h5'
        keras_model_json = model + '.json'
        keras_model = model_from_json(open(keras_model_json).read())
        keras_model.load_weights(keras_model_weights)
        return keras_model, keras_model_weights, keras_model_json

    model, model_weights, model_json = load_model(os.path.join(py_dir,'example_models/%s/deepstarr.model' % model_name))
    """
    Inputs shape:   (n, 249, 4)
    Outputs shape:  (2,)
    """

    # define sequence-of-interest and output head (task)
    seq_index = 24869
    task_idx = 0 #developmental (DEV) class

    # create save directory
    if save is True:
        save_dir = os.path.join(py_dir, 'squid_outputs/%s/%s' % (model_name, seq_index))
        if not os.path.exists(save_dir):
            os.makedirs(save_dir)
    else:
        save_dir = None

    # retrieve genomic sequences from test set
    import h5py
    with h5py.File(os.path.join(py_dir, 'example_models/%s/deepstarr_data.h5' % model_name), 'r') as dataset:
        x_all = np.array(dataset['x_test']).astype(np.float32)

   # define mutagenesis window for sequence
    start_position = 123-10 #e.g., the nucleotide at position 123 is the start of a pronounced GATA motif (GATAA) in the DEV program
    stop_position = 123+5+10

    # set up predictor class for in silico MAVE
    pred_generator = predictor.ScalarPredictor(pred_fun=model.predict_on_batch, 
                                               task_idx=task_idx, batch_size=512)
    
    alphabet = ['A','C','G','T']
    log2FC = False
    output_skip = 0

else:
    # empty section for user-defined model
    pass


# =============================================================================
# Setup for workflow independent from choice of deep learning model
# =============================================================================
x = x_all[seq_index]
seq_length = x.shape[0]
mut_window = [start_position, stop_position]


# set up mutagenizer class for in silico MAVE
mut_generator = mutagenizer.RandomMutagenesis(mut_rate=0.1, uniform=False)

# generate in silico MAVE
mave = mave.InSilicoMAVE(mut_generator, pred_generator, seq_length, mut_window=mut_window, log2FC=log2FC)
x_mut, y_mut = mave.generate(x, num_sim=10000, seed=None)

# set up surrogate model
gpmap = 'additive' #options: {'additive', 'pairwise' if MAVE-NN}
if 1:
    surrogate_model = surrogate_zoo.SurrogateMAVENN(x_mut.shape, num_tasks=y_mut.shape[1],
                                                    gpmap=gpmap, regression_type='GE',
                                                    linearity='nonlinear', noise='SkewedT',
                                                    noise_order=2, reg_strength=0.1,
                                                    alphabet=alphabet, deduplicate=True, gpu=gpu)
else:
    surrogate_model = surrogate_zoo.SurrogateLinear(x_mut.shape, num_tasks=y_mut.shape[1],
                                                    alphabet=alphabet, gpu=gpu)

# train surrogate model
surrogate, mave_df = surrogate_model.train(x_mut, y_mut, learning_rate=5e-4, epochs=500, batch_size=100,
                                           early_stopping=True, patience=25, restore_best_weights=True,
                                           save_dir=save_dir, verbose=1)

# retrieve model parameters
params = surrogate_model.get_params(gauge='empirical', save_dir=save_dir)

# generate sequence logo
logo = surrogate_model.get_logo(mut_window=mut_window, full_length=seq_length)

if mave_df is not None: #below set up for MAVENN models only
    # retrieve model performance metrics
    info = surrogate_model.get_info(save_dir=save_dir, verbose=True)

    # plot figures
    impress.plot_y_hist(y_mut, save_dir=save_dir)
    impress.plot_performance(surrogate, info=info, save_dir=save_dir) #plot model performance (bits)
    impress.plot_additive_logo(logo, center=True, view_window=mut_window, alphabet=alphabet, save_dir=save_dir)
    if gpmap == 'pairwise':
        impress.plot_pairwise_matrix(params[2], view_window=mut_window, alphabet=alphabet, save_dir=save_dir)
    impress.plot_y_vs_yhat(surrogate, mave_df=mave_df, save_dir=save_dir)
    impress.plot_y_vs_phi(surrogate, mave_df=mave_df, save_dir=save_dir)



# Parameters to include above:
'''
output_skip = 0 #ENFORMER
pred_trans_delimit = None #ENFORMER
max_in_mem = False #ENFORMER
'''


"""
Library of functions for DNN inference on mutagenized sequences
"""

import os, sys
sys.dont_write_bytecode = True
import numpy as np
#import impress


class BasePredictor():

    def __init__(self, pred_fun, reduce_fun, task_idx, batch_size):
        self.pred_fun = pred_fun
        self.reduce_fun = reduce_fun
        self.task_idx = task_idx
        self.batch_size = batch_size

    def __call__(self, x):
        """Return an in silico MAVE based on mutagenesis of 'x'.

        Parameters
        ----------
        x : torch.Tensor
            Batch of one-hot sequences (shape: (L, A)).

        Returns
        -------
        torch.Tensor
            Batch of one-hot sequences with random augmentation applied.
        """
        raise NotImplementedError()



class ScalarPredictor(BasePredictor):

    def __init__(self, pred_fun, task_idx=0, batch_size=64, **kwargs):
        self.pred_fun = pred_fun
        self.task_idx = task_idx
        self.kwargs = kwargs
        self.batch_size = batch_size

    def __call__(self, x):
        pred = prediction_in_batches(x, self.pred_fun, self.batch_size, **self.kwargs)
        #return pred[:,self.task_idx][:,np.newaxis]
        return pred[self.task_idx]



class ProfilePredictor(BasePredictor):

    def __init__(self, pred_fun, task_idx=0, batch_size=64, reduce_fun=np.sum, save_dir=None, **kwargs):
        self.pred_fun = pred_fun
        self.task_idx = task_idx
        self.batch_size = batch_size
        self.reduce_fun = reduce_fun
        #self.axis = axis
        self.save_dir = save_dir
        self.kwargs = kwargs

    def __call__(self, x):
        # get model predictions (all tasks)
        pred = prediction_in_batches(x, self.pred_fun, self.batch_size, **self.kwargs)

        # reduce profile to scalar across axis for a given task_idx
        pred = self.reduce_fun(pred[:,:,self.task_idx], save_dir=self.save_dir)
        return pred[:,np.newaxis]



class BPNetPredictor(BasePredictor):

    def __init__(self, pred_fun, task_idx=0, batch_size=64, reduce_fun=np.sum, axis=1, strand='pos', **kwargs):
        self.pred_fun = pred_fun
        self.task_idx = task_idx
        self.batch_size = batch_size
        self.reduce_fun = reduce_fun
        self.axis = axis
        self.kwargs = kwargs
        if strand == 'pos':
            self.strand = 0
        else:
            self.strand = 1

    def __call__(self, x):

        # get model predictions (all tasks)
        pred = prediction_in_batches(x, self.pred_fun, self.batch_size, **self.kwargs)

        # reduce bpnet profile prediction to scalar across axis for a given task_idx
        pred = pred[self.task_idx][0][:,self.strand]
        pred = self.reduce_fun(pred, axis=self.axis)
        return pred[:,np.newaxis]




################################################################################
# useful functions
################################################################################


def prediction_in_batches(x, model_pred_fun, batch_size=None, **kwargs):

    N, L, A = x.shape
    num_batches = np.floor(N/batch_size).astype(int)
    pred = []
    for i in range(num_batches):
        pred.append(model_pred_fun(x[i*batch_size:(i+1)*batch_size], **kwargs))
    if num_batches*batch_size < N:
        pred.append(model_pred_fun(x[num_batches*batch_size:], **kwargs))

    try:
        preds = np.concatenate(pred, axis=1)
    except ValueError as ve:
        preds = np.vstack(pred)
    return preds



def profile_sum(pred, save_dir=None):

    sum = np.sum(pred, axis=1)
    return sum


def profile_pca(pred, save_dir=None):

    N, B = pred.shape #B : number of bins in profile
    Y = pred.copy()
    sum = np.sum(pred, axis=1) #needed for sense correction

    # normalization: mean of all distributions is subtracted from each distribution
    mean_all = np.mean(Y, axis=0)
    for i in range(N):
        Y[i,:] -= mean_all

    u,s,v = np.linalg.svd(Y.T, full_matrices=False)
    vals = s**2 #eigenvalues
    vecs = u #eigenvectors
    
    U = Y.dot(vecs)
    v1, v2 = 0, 1
    
    corr = np.corrcoef(sum, U[:,v1])
    if corr[0,1] < 0: #correct for eigenvector "sense"
        U[:,v1] = -1.*U[:,v1]

    #impress.plot_eig_vals(vals, save_dir=save_dir)
    #impress.plot_eig_vecs(U, v1=v1, v2=v2, save_dir=save_dir)

    return U[:,v1]


"""
def custom_reduce(pred):
    # code to reduce predictions to (N,1)
    return pred_reduce
"""


import os, sys
sys.dont_write_bytecode = True
import pandas as pd
import numpy as np



def arr2pd(x, letters=['A','C','G','T']):
    """Convert Numpy array to Pandas dataframe with proper column headings.

    Parameters
    ----------
    x : ARRAY with shape (L, 4)
        input sequence (one-hot encoding or attribution map).
    letters : 1D ARRAY
        All characters present in the sequence alphabet (e.g., ['A','C','G','T'] for DNA)

    Returns
    -------
    x : DATAFRAME
        Pandas dataframe corresponding to the input Numpy array
    """
    
    labels = {}
    idx = 0
    for i in letters:
        labels[i] = x[:,idx]
        idx += 1
    x = pd.DataFrame.from_dict(labels, orient='index').T
    
    return x


def oh2seq(OH, alphabet):
    """
    Convert one-hot encoding to sequence
    ----------
    OH : ARRAY with shape (L, 4)
        Input sequence (one-hot encoding)
    alphabet : 1D ARRAY
        All characters present in the sequence alphabet (e.g., ['A','C','G','T'] for DNA)
    
    Returns
    -------
    seq : STRING with length L
        Sequence corresponding to input one-hot encoding (e.g., 'AATGAC...')
    """
    
    seq = []
    for i in range(np.shape(OH)[0]):
        for j in range(len(alphabet)):
            if OH[i][j] == 1:
                seq.append(alphabet[j])
    seq = ''.join(seq)
    return seq


def seq2oh(seq, alphabet):
    """
    Convert sequence to one-hot encoding
    ----------
    seq : STRING with length L
        Input sequence
    alphabet : 1D ARRAY
        All characters present in the sequence alphabet (e.g., ['A','C','G','T'] for DNA)
    
    Returns
    -------
    OH : ARRAY with shape (L, 4)
        One-hot encoding corresponding to input sequence
    """
    
    L = len(seq)
    OH = np.zeros(shape=(L,len(alphabet)), dtype=np.float32)
    for idx, i in enumerate(seq):
        for jdx, j in enumerate(alphabet):
            if i == j:
                OH[idx,jdx] = 1
    return OH


def fix_gauge(x, gauge, wt=None, r=None):
    """    
    Fix the gauge for an attribution matrix
    
    x :         ARRAY with shape (L, 4)
                Matrix of attribution scores for a sequence-of-interest
    gauge :     STRING {'empirical', 'wildtype', 'hierarchical', 'default'}
                Specification of which gauge to use
    OH_wt :     ARRAY with shape (L, 4)
                Wild-type sequence (one-hot encoding); needed if gauge = 'wildtype'
    r :         Probability of mutation used during generation of in silico MAVE dataset

    Returns
    -------
    OH : ARRAY with shape (L, 4)
        Gauge-fixed one-hot encoding corresponding to input sequence
    """

    x1 = x.copy()

    if gauge == 'empirical':
        L = wt.shape[0] #length of sequence
        wt_argmax = np.argmax(wt, axis=1) #index of each wild-type in the one-hot encoding

        p_lc = np.ones(shape=wt.shape) #empirical probability matrix
        p_lc = p_lc*(r/3.)

        for l in range(L):
            p_lc[l,wt_argmax[l]] = (1-r)

        for l in range(L):
            weighted_avg = np.average(x[l,:], weights=p_lc[l,:])
            for c in range(4):
                x1[l,c] -= weighted_avg

    elif gauge == 'wildtype':
        L = wt.shape[0]
        wt_argmax = np.argmax(wt, axis=1)
        for l in range(L):
            wt_val = x[l, wt_argmax[l]]
            x1[l,:] -= wt_val

    elif gauge == 'hierarchical':
        for l in range(x.shape[0]):
            col_mean = np.mean(x[l,:])
            x1[l,:] -= col_mean

    elif gauge == 'default':
        pass

    return x1


"""
Functions for plotting intermediate and final data
"""

import os, sys
sys.dont_write_bytecode = True
import numpy as np
import matplotlib.pyplot as plt
from matplotlib.ticker import MaxNLocator
import warnings
warnings.simplefilter(action='ignore', category=FutureWarning)
import mavenn
import logomaker
from . import utils


def plot_y_hist(y_mut, save_dir):

    # plot histogram of transformed deepnet predictions
    fig, ax = plt.subplots()
    ax.hist(y_mut, bins=100)
    ax.set_xlabel('y')
    ax.set_ylabel('Frequency')
    ax.axvline(y_mut[0], c='red', label='WT', linewidth=2, zorder=10) #wild-type prediction
    plt.legend(loc='upper right')
    plt.tight_layout()
    if save_dir is not None:
        plt.savefig(os.path.join(save_dir, 'mave_distribution.png'), facecolor='w', dpi=200)
        plt.close()
    else:
        plt.show()


def plot_performance(model, info, save_dir):
    
    # plot mavenn model performance
    fig, ax = plt.subplots(1, 1, figsize=[5, 5])
    # plot I_var_train, the variational information on training data as a function of epoch
    ax.plot(model.history['I_var'], label=r'I_var_train')
    # plot I_var_val, the variational information on validation data as a function of epoch
    ax.plot(model.history['val_I_var'], label=r'val_I_var')
    # plot I_pred_test, the predictive information of the final model on test data
    ax.axhline(info, color='C3', linestyle=':', label=r'test_I_pred')
    ax.set_xlabel('epochs')
    ax.set_ylabel('bits')
    ax.set_title('Training history')
    ax.yaxis.set_major_locator(MaxNLocator(integer=True))
    ax.legend(loc='best')
    plt.tight_layout()
    if save_dir is not None:
        plt.savefig(os.path.join(save_dir, 'mavenn_training.png'), facecolor='w', dpi=200)
        plt.close()
    else:
        plt.show()


def plot_additive_logo(logo, center=True, view_window=None, alphabet=['A','C','G','T'], fig_size=None, save_dir=None):

    # plot additive logo
    if fig_size is not None:
        fig, ax = plt.subplots(figsize=fig_size)
    else:
        fig, ax = plt.subplots(figsize=[10,3])

    if view_window is None:
        logo_fig = logo
    else:
        logo_fig = logo[view_window[0]:view_window[1]]
        logo_fig = utils.arr2pd(logo_fig, alphabet)

    logomaker.Logo(df=logo_fig,
                    ax=ax,
                    fade_below=.5,
                    shade_below=.5,
                    width=.9,
                    center_values=center,
                    font_name='Arial Rounded MT Bold')
    if view_window is not None:
        ax.set_xticks(np.arange(0, view_window[1]-view_window[0], 1))
        ax.set_xticklabels(np.arange(view_window[0], view_window[1], 1))

    ax.xaxis.set_major_locator(MaxNLocator(integer=True))
    ax.set_ylabel('Additive effect')
    ax.set_xlabel('Nucleotide position')
    plt.tight_layout()
    if save_dir is not None:
        plt.savefig(os.path.join(save_dir, 'additive_logo.png'), facecolor='w', dpi=200)
        plt.close()
    else:
        plt.show()


def plot_pairwise_matrix(theta_lclc, view_window=None, alphabet=['A','C','G','T'], save_dir=None):

    # plot maveen pairwise matrix
    fig, ax = plt.subplots(figsize=[10,5])
    ax, cb = mavenn.heatmap_pairwise(values=theta_lclc,
                                    alphabet=alphabet,
                                    ax=ax,
                                    gpmap_type='pairwise',
                                    cmap_size='2%',
                                    show_alphabet=False,
                                    cmap='seismic',
                                    cmap_pad=.1,
                                    show_seplines=True,            
                                    sepline_kwargs = {'color': 'k',
                                                      'linestyle': '-',
                                                      'linewidth': .5})
    if view_window is not None:
        ax.xaxis.set_ticks(np.arange(0, view_window[1]-view_window[0], 2))
        ax.set_xticklabels(np.arange(view_window[0], view_window[1], 2))  
    cb.set_label(r'Pairwise Effect',
                  labelpad=8, ha='center', va='center', rotation=-90)
    cb.outline.set_visible(False)
    cb.ax.tick_params(direction='in', size=20, color='white')
    ax.set_xlabel('Nucleotide position')
    plt.tight_layout()
    if save_dir is not None:
        plt.savefig(os.path.join(save_dir, 'pairwise_matrix.png'), facecolor='w', dpi=200)
        plt.close()
    else:
        plt.show()


def plot_y_vs_yhat(model, mave_df, save_dir=None):

    # plot mavenn y versus yhat
    fig, ax = plt.subplots(1,1,figsize=[5,5])
    trainval_df, test_df = mavenn.split_dataset(mave_df)
    y_test = test_df['y'] #get test data y values
    yhat_test = model.x_to_yhat(test_df['x']) #compute yhat on test data
    Rsq = np.corrcoef(yhat_test.ravel(), test_df['y'])[0, 1]**2 #compute R^2 between yhat_test and y_test    
    ax.scatter(yhat_test, y_test, color='C0', s=1, alpha=.1,
               label='test data')
    xlim = [min(yhat_test), max(yhat_test)]
    ax.plot(xlim, xlim, '--', color='k', label='diagonal', zorder=100)
    ax.set_xlabel('model prediction ($\hat{y}$)')
    ax.set_ylabel('measurement ($y$)')
    ax.set_title(f'Standard metric of model performance:\n$R^2$={Rsq:.3}');
    ax.legend(loc='upper left')
    plt.tight_layout()
    if save_dir is not None:
        plt.savefig(os.path.join(save_dir,'mavenn_measure_yhat.png'), facecolor='w', dpi=200)
        plt.close()
    else:
        plt.show()
    
    
def plot_y_vs_phi(model, mave_df, save_dir=None):

    # plot mavenn y versus phi
    fig, ax = plt.subplots(1,1,figsize=[5,5])
    trainval_df, test_df = mavenn.split_dataset(mave_df)
    phi_test = model.x_to_phi(test_df['x']) #compute φ on test data
    phi_lim = [min(phi_test)-.5, max(phi_test)+.5] #set phi limit and create a grid in phi space
    phi_grid = np.linspace(phi_lim[0], phi_lim[1], 1000)
    yhat_grid = model.phi_to_yhat(phi_grid) #compute yhat each phi grid point
    q = [0.025, 0.975] #compute 95% CI for each yhat
    yqs_grid = model.yhat_to_yq(yhat_grid, q=q)
    ax.fill_between(phi_grid, yqs_grid[:, 0], yqs_grid[:, 1],
                   alpha=0.1, color='C1', lw=0, label='95% CI') #plot 95% confidence interval
    ax.plot(phi_grid, yhat_grid,
            linewidth=2, color='C1', label='nonlinearity') #plot GE nonlinearity
    y_test = test_df['y']
    ax.scatter(phi_test, y_test,
               color='C0', s=1, alpha=.1, label='test data',
               zorder=-100, rasterized=True) #plot scatter of φ and y values
    ax.set_xlim(phi_lim)
    ax.set_xlabel('latent phenotype ($\phi$)')
    ax.set_ylabel('measurement ($y$)')
    ax.set_title('GE measurement process')
    ax.legend(loc='upper left')
    fig.tight_layout()
    if save_dir is not None:
        plt.savefig(os.path.join(save_dir, 'mavenn_measure_phi.png'), facecolor='w', dpi=200)
        plt.close()
    else:
        plt.show()


def plot_eig_vals(vals, save_dir=None):

    x = range(1,len(vals)+1)
    plt.scatter(x, vals)
    plt.title('Eigenvalue spectrum')
    plt.xlabel(r'$PC$')
    plt.ylabel(r'$\mathrm{\lambda}$', rotation=0)
    plt.ticklabel_format(style='sci', axis='y', scilimits=(0,0))
    plt.xlim([0,15])
    plt.ylim(-vals[0]/8., vals[0]+vals[0]/8.)
    plt.locator_params(nbins=15)
    plt.axhline(y=0, color='k', alpha=.5, linestyle='--', linewidth=1)
    if save_dir is not None:
        plt.savefig(os.path.join(save_dir, 'pca_eigvals.png'), facecolor='w', dpi=200)
        plt.close()
    else:
        plt.show()


def plot_eig_vecs(U, v1, v2, save_dir=None):

    fig, ax = plt.subplots()
    ax.scatter(U[:,v1], U[:,v2], s=1, facecolor='k', alpha=.5)
    ax.scatter(U[:,v1][0], U[:,v2][0], s=5, facecolor='r', label='WT') #first index is wild-type
    ax.set_xlabel(r'$PC_%s$' % (v1+1), fontsize=20)
    ax.set_ylabel(r'$PC_%s$' % (v2+1), fontsize=20)
    plt.legend(loc='best')
    plt.tight_layout()
    if save_dir is not None:
        plt.savefig(os.path.join(save_dir, 'pca_eigvecs.png'), facecolor='w', dpi=200)
        plt.close()
    else:
        plt.show()


"""
Functions for performing in silico MAVE
"""

import os, sys
sys.dont_write_bytecode = True
import numpy as np


class InSilicoMAVE():
    """Module for performing in silico MAVE.

    Parameters
    ----------
    mut_rate : float, optional
        Mutation rate for random mutagenesis (defaults to 0.1).
    start_position : int, optional
        Index of start position along sequence to probe (defaults to 0).
    stop_position : int, optional
        Index of stop position along sequence to probe (defaults to None).
    uniform : bool
        uniform (True), Poisson (false); sets the number of mutations per sequence
    """
    def __init__(self, mut_generator, mut_predictor, seq_length, mut_window=None, log2FC=False):
        self.mut_generator = mut_generator
        self.mut_predictor = mut_predictor
        self.seq_length = seq_length
        self.mut_window = mut_window
        self.log2FC = log2FC
        if mut_window is not None:
            self.start_position = mut_window[0]
            self.stop_position = mut_window[1]
        else:
            self.start_position = 0
            self.stop_position = seq_length


    def generate(self, x, num_sim, seed=None, verbose=1):
        """Randomly mutate segments in a set of one-hot DNA sequences.

        Parameters
        ----------
        x : torch.Tensor
            Batch of one-hot sequences (shape: (L,A)).
        num_sim : int
            Number of sequences to mutagenize for in silico MAVE.
        seed : int, optional
            sets the random number seed

        Returns
        -------
        torch.Tensor
            Sequences with randomly mutated segments (padded to correct shape
            with random DNA)
        """
        np.random.seed(seed)
        if verbose:
            print('')
            print('Building in silico MAVE...')

        # generate in silico MAVE based on mutagenesis strategy
        if verbose:
            print('  Generating mutagenized sequences...')
        if self.mut_window is not None:
            x_window = self.delimit_range(x, self.start_position, self.stop_position)
            x_mut = self.mut_generator(x_window, num_sim)
        else:
            x_mut = self.mut_generator(x, num_sim)

        # predict MAVE data and process
        if verbose:
            print('  Predicting mutagenized sequences...')
        if self.mut_window is not None:
            x_mut_full = self.pad_mave(x_mut, x, self.start_position, self.stop_position)
            y_mut = self.mut_predictor(x_mut_full)
        else:
            y_mut = self.mut_predictor(x_mut)

        if self.log2FC is True:
            y_mut = self.apply_log2FC(y_mut)

        return x_mut, y_mut


    def pad_mave(self, x_mut, x, start_position, stop_position):
        N = x_mut.shape[0]
        x = x[np.newaxis,:]
        x_start = np.tile(x[:,:start_position,:], (N,1,1))
        x_stop = np.tile(x[:,stop_position:,:], (N,1,1))
        return np.concatenate([x_start, x_mut, x_stop], axis=1)


    def delimit_range(self, x, start_position, stop_position):
        return x[start_position:stop_position,:]
    

    def apply_log2FC(self, y):

        if np.amin(y) < 0:
            y += (abs(np.amin(y)) + 1)
        elif 0 <= np.amin(y) < 1:
            y += 1

        y_log2_wt = np.log2(y[0])
        y_log2_all = np.log2(y)
        y_log2_fc = y_log2_wt - y_log2_all

        return y_log2_fc